#!/bin/bash

#out_dir=./
out_dir=../cubin/
#out_dir=/home/basy/AtomicNetworksCVS/Server/Processor/cubin/

maxrregcount=16

job="/usr/local/cuda/bin/nvcc -m64 -DUNIX -O3 --compiler-options -fno-strict-aliasing -Xptxas -v "
if [[ $maxrregcount > 0 ]] ; then job=$job"--maxrregcount=$maxrregcount " ; fi

rm "$out_dir"*

colors=1
if [[ $# -ge 1 && "$1" == "-nc" || "$1" == "--no-color" ]] ; then colors=0 ; fi

source /home/basy/scripts/bashcolors

tabs 1 25 60 70 79 85 94 

for f in $(ls *.cu)
do

  if [[ $colors -eq 1 ]] ; then
    echo -e "${magenta}Building kernel${NC}[${WHITE}$f${NC}]"
  else
    echo "Building kernel[$f]"
  fi

  ptx="$(basename $f .cu).ptx"
  $job -cubin ./$f -o "$out_dir""$f"bin > "$out_dir""$ptx" 2>&1
  if [[ $(cat "$out_dir""$ptx") == *error* ]] ; then 
    cat "$out_dir""$ptx"
    if [[ $colors -eq 1 ]] ; then
      echo -e "${red}Error in file ${RED}$f${red}, check in ${RED}$out_dir$ptx${red} !!!${NC}"
    else
      echo "Error in file[$f], check in $out_dir$ptx !!!"
    fi
    exit 1
  fi
  #cat "$out_dir""$ptx"

  
  if [[ $colors -eq 1 ]] ; then
    while read line
    do
      if [[ $( echo "$line" | grep -i warning ) ]] ; then 
        echo -e "${red}$line${NC}"
        continue
      fi
      if [[ -z $FUNC_NAME ]] ; then
        FUNC_NAME=$(echo $line | awk '{print $7}')
        FUNC_NAME=${FUNC_NAME:1:$(( ${#FUNC_NAME} - 2 ))}
      else
        regs=$(echo $line | egrep ".*Used [0-9]* registers"  | sed -e "s/.*Used \([0-9]*\) registers.*/\1/")
        smem=$(echo $line | egrep "[0-9]*+[0-9]* bytes smem" | sed -e "s/.* \([0-9]*\)+\([0-9]*\) bytes smem.*/\1\t\2/")
        lmem=$(echo $line | egrep "[0-9]*+[0-9]* bytes lmem" | sed -e "s/.* \([0-9]*\)+\([0-9]*\) bytes lmem.*/\1\t\2/")
        if [[ ! $smem ]] ; then smem="0\t0"; fi
        if [[ ! $lmem ]] ; then lmem="0\t0"; fi
        echo -en "${magenta}Kernel${NC}[${WHITE}$f${NC}]\t"
        echo -en "${GREEN}Function${NC}[${YELLOW}$FUNC_NAME${NC}]\t"
        echo -en "${cyan}regs${NC}[${CYAN}$regs${NC}]\t"
        echo -en "${blue}smem${NC}[${BLUE}$smem${NC}]\t"
        echo -e  "${blue}lmem${NC}[${BLUE}$lmem${NC}]"
        if [[ $regs ]] ; then
          echo "$regs" > $out_dir""$(basename $f .cu)"_"$FUNC_NAME".regs"
        fi
        FUNC_NAME=""
      fi
    done < "$out_dir""$ptx"
  else
    cat "$out_dir""$ptx"
    echo
  fi
done

exit 0

