#ifndef ___PROCES_F_CELL_STRESS_BASE_H___
#define ___PROCES_F_CELL_STRESS_BASE_H___

#include "proces_f_cell_data.h"

__device__ float StressCelluleBase(float curSize, float3 * newPosMatrix, float3 * matePos, float3 * mateOptPos){
  float tmpSize;
  float3 tmp;
  tmp.x = newPosMatrix->x + ( matePos->x - mateOptPos->x );
  tmp.y = newPosMatrix->y + ( matePos->y - mateOptPos->y );
  tmp.z = newPosMatrix->z + ( matePos->z - mateOptPos->z );
  tmpSize = sqrt(tmp.x*tmp.x + tmp.y*tmp.y + tmp.z*tmp.z);
  if(tmpSize > curSize){
    newPosMatrix->x = tmp.x;
    newPosMatrix->y = tmp.y;
    newPosMatrix->z = tmp.z;
    return tmpSize;
  }
  return curSize;
}

////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

__device__ float StressCelluleMateBase( float3 * posIn, CellAttribute * attr, CrystalStructure crStr,
                                        int dir, int mateId, int nearInterstitial, int curDataId, int dataCount,
                                        short curTempr, unsigned long curMask, float curSize,
                                        float3 * curPos, float3 * newPosMatrix){

  int           mateType, mateDataId;
  unsigned long mateMask;
  float3        matePos, mateOptPos;

  mateType  = attr[mateId].type;
  mateMask  = attr[mateId].mask;
  __syncthreads();

  matePos.x = posIn[mateId].x;
  matePos.y = posIn[mateId].y;
  matePos.z = posIn[mateId].z;
  __syncthreads();

  if(__IS_CELL_MASK_BUG_INT_ANY(curMask))return curSize;
  __syncthreads();

  if(nearInterstitial != -1) {
    if(!__IS_CELL_MASK_BUG_INT_ANY(mateMask)){
      if(crStr == BCC__CrystalStructure){
        SetLaticePositionsBCC2(&mateOptPos, curPos, dir, nearInterstitial);
        return StressCelluleBase(curSize, newPosMatrix, &matePos, &mateOptPos);
      }
      if(crStr == FCC__CrystalStructure){
        SetLaticePositionsFCC2(&mateOptPos, curPos, dir, nearInterstitial);
        return StressCelluleBase(curSize, newPosMatrix, &matePos, &mateOptPos);
      }
      return 0.0f;
    }
    else return 0.0f;
  }
  __syncthreads();

  if(__IS_CELL_MASK_NEAR_BUG(curMask)){
    if(__IS_CELL_MASK_BUG_VAC(mateMask))        mateDataId = GetCalcDataIDX(true,  curTempr, cellData[curDataId].id, -1, -1, crStr, dataCount);
    else if(__IS_CELL_MASK_BUG_SUB(mateMask))   mateDataId = GetCalcDataIDX(false, curTempr, cellData[curDataId].id, mateType, -1, crStr, dataCount);
    else                                        mateDataId = curDataId;
    __syncthreads();
    if(mateDataId != -1){
      if(crStr == BCC__CrystalStructure){
        SetLaticePositionsBCC2(&mateOptPos, curPos, dir, mateDataId);
        return StressCelluleBase(curSize, newPosMatrix, &matePos, &mateOptPos);
      }
      __syncthreads();
      if(crStr == FCC__CrystalStructure){
        SetLaticePositionsFCC2(&mateOptPos, curPos, dir, mateDataId);
        return StressCelluleBase(curSize, newPosMatrix, &matePos, &mateOptPos);
      }
      __syncthreads();
      return curSize;
    }
    else return curSize;
  }
  __syncthreads();

  if(crStr == BCC__CrystalStructure){
    SetLaticePositionsBCC2(&mateOptPos, curPos, dir, curDataId);
    return StressCelluleBase(curSize, newPosMatrix, &matePos, &mateOptPos);
  }

  if(crStr == FCC__CrystalStructure){
    SetLaticePositionsBCC2(&mateOptPos, curPos, dir, curDataId);
    return StressCelluleBase(curSize, newPosMatrix, &matePos, &mateOptPos);
  }
  __syncthreads();
  return curSize;
}

////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

extern "C" __global__ void StressCellulesBase(float3 * stressOut, float3 * posIn, CellAttribute * attr, CellMate * mate, short * tempr, CalcDataF * data,
                                              int cellCount, int dataCount, int steps){

  int idx = blockIdx.x * blockDim.x + threadIdx.x;
  int i, j;
  int cntLoops, cntMates;
  int mateId, curDataId, nearInterstitial;
  short curTempr;
  CrystalStructure crStr;
  CellAttribute curAttr;
  float  curStress;
  float3 curPos, newPosMatrix;

  if(idx < dataCount)StoreCalcDataF(idx, data);
  __syncthreads();

  for(i = idx * steps ; (i < cellCount && i < (idx + 1) * steps); i++){
    curAttr.type      = attr[i].type;
    curAttr.grainIdx  = attr[i].grainIdx;
    curAttr.mask      = attr[i].mask;
    __syncthreads();
    curTempr          = tempr[i];
    __syncthreads();

    crStr = CellMaskGetCrystalStruct(curAttr.mask);

    curDataId = GetCalcDataIDX(__IS_CELL_MASK_BUG_VAC(curAttr.mask), curTempr, curAttr.grainIdx,
                               __IS_CELL_MASK_BUG_SUB(curAttr.mask)      ? curAttr.type : -1,
                               __IS_CELL_MASK_BUG_INT_ANY(curAttr.mask)  ? curAttr.type : -1,
                               crStr, dataCount);
    __syncthreads();

    if(curDataId != -1 && curAttr.mask != CELL_MASK_NULL && !__IS_CELL_MASK_SCENE_BORDER(curAttr.mask)){

      curStress       = 0.0f;
      newPosMatrix.x  = 0.0f; newPosMatrix.y  = 0.0f; newPosMatrix.z  = 0.0f;
      __syncthreads();

      curPos.x = posIn[i].x;
      curPos.y = posIn[i].y;
      curPos.z = posIn[i].z;
      __syncthreads();

      if(__IS_CELL_MASK_BUG_INT_ANY(curAttr.mask)){ cntLoops = 0; }
      else if(crStr == BCC__CrystalStructure){      cntLoops = CNT_DIR_BCC; }
      else if(crStr == FCC__CrystalStructure){      cntLoops = CNT_DIR_FCC; }
      else{                                         cntLoops = 0; }
      __syncthreads();

      nearInterstitial = -1;
      if(__IS_CELL_MASK_NEAR_BUG(curAttr.mask)){
        for(j = 0 ; j < cntLoops ; j++){
          mateId = mate[i].m1.mate[j];                                                                                        __syncthreads();
          if( mateId != 0)nearInterstitial = IsNearInterstitial(attr, mateId, curTempr, curAttr.grainIdx, crStr, dataCount);  __syncthreads();
          if(-1 != nearInterstitial)break;                                                                                    __syncthreads();
          mateId = mate[i].m2.mate[j];                                                                                        __syncthreads();
          if( mateId != 0)nearInterstitial = IsNearInterstitial(attr, mateId, curTempr, curAttr.grainIdx, crStr, dataCount);  __syncthreads();
          if(-1 != nearInterstitial)break;                                                                                    __syncthreads();
          mateId = mate[i].m3.mate[j];                                                                                        __syncthreads();
          if( mateId != 0)nearInterstitial = IsNearInterstitial(attr, mateId, curTempr, curAttr.grainIdx, crStr, dataCount);  __syncthreads();
          if(-1 != nearInterstitial)break;                                                                                    __syncthreads();
        }
        __syncthreads();
      }
      __syncthreads();

      for(j = 0 ; j < cntLoops ; j++){
        cntMates = CellMaskGetCntMates(curAttr.mask, j);
        __syncthreads();

        if(cntMates > 0){
          mateId = mate[i].m1.mate[j];  __syncthreads();
          if( mateId != 0){
            curStress = StressCelluleMateBase(posIn, attr, crStr, j, mateId, nearInterstitial, curDataId, dataCount, curTempr, curAttr.mask, curStress, &curPos, &newPosMatrix);
            __syncthreads();
          }
        }
        __syncthreads();

        if(cntMates > 1){
          mateId = mate[i].m2.mate[j]; __syncthreads();
          if( mateId != 0){
            curStress = StressCelluleMateBase(posIn, attr, crStr, j, mateId, nearInterstitial, curDataId, dataCount, curTempr, curAttr.mask, curStress, &curPos, &newPosMatrix);
            __syncthreads();
          }
        }
        __syncthreads();

        if(cntMates > 2){
          mateId = mate[i].m3.mate[j]; __syncthreads();
          if( mateId != 0){
            curStress = StressCelluleMateBase(posIn, attr, crStr, j, mateId, nearInterstitial, curDataId, dataCount, curTempr, curAttr.mask, curStress, &curPos, &newPosMatrix);
            __syncthreads();
          }
        }
        __syncthreads();
      }
      __syncthreads();
    }
    __syncthreads();

    stressOut[i].x = newPosMatrix.x;
    stressOut[i].y = newPosMatrix.y;
    stressOut[i].z = newPosMatrix.z;
    __syncthreads();

  }
  __syncthreads();
}

#endif
