#ifndef ___PROCES_F_EXT_STABIL_H___
#define ___PROCES_F_EXT_STABIL_H___

#ifndef CUDA_KERNEL_INCLUDE
  #define CUDA_KERNEL_INCLUDE
  #define QT_VERSION
#endif

#include "../cellmask.h"

extern "C" __global__ void FastBorderStabilX(float3 * out, ulong * mask, int steps, int cellCount, float sceneSizeX, float farX, float moveX){
  int idx = (blockIdx.x * blockDim.x + threadIdx.x) * steps;
  float tmp;
  for(int i = 0 ; i < steps ; i++){
    if(idx < cellCount && mask[idx] != CELL_MASK_NULL){
      tmp = out[idx].x;                                 __syncthreads();
      tmp = tmp + ((tmp - farX) / sceneSizeX) * moveX;  __syncthreads();
      out[idx].x = tmp;                                 __syncthreads();
    }
    idx++;
    __syncthreads();
  }
  __syncthreads();
}

////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

extern "C" __global__ void FastBorderStabilY(float3 * out, ulong * mask, int steps, int cellCount, float sceneSizeY, float farY, float moveY){
  int idx = (blockIdx.x * blockDim.x + threadIdx.x) * steps;
  float tmp;
  for(int i = 0 ; i < steps ; i++){
    if(idx < cellCount && mask[idx] != CELL_MASK_NULL){
      tmp = out[idx].y;                                 __syncthreads();
      tmp = tmp + ((tmp - farY) / sceneSizeY) * moveY;  __syncthreads();
      out[idx].y = tmp;                                 __syncthreads();
    }
    idx++;
    __syncthreads();
  }
  __syncthreads();
}

////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

extern "C" __global__ void FastBorderStabilZ(float3 * out, ulong * mask, int steps, int cellCount, float sceneSizeZ, float farZ, float moveZ){
  int idx = (blockIdx.x * blockDim.x + threadIdx.x) * steps;
  float tmp;
  for(int i = 0 ; i < steps ; i++){
    if(idx < cellCount && mask[idx] != CELL_MASK_NULL){
      tmp = out[idx].z;                                 __syncthreads();
      tmp = tmp + ((tmp - farZ) / sceneSizeZ) * moveZ;  __syncthreads();
      out[idx].z = tmp;                                 __syncthreads();
    }
    idx++;
    __syncthreads();
  }
  __syncthreads();
}

#endif
