#ifndef ___PROCES_F_STABIL_BASE_H___
#define ___PROCES_F_STABIL_BASE_H___

#include "proces_f_cell_data.h"

__device__ void StabilCelluleBase(float curMateCnt, float3 * newPosMatrix, float3 * matePos, float3 * mateOptPos){
  newPosMatrix->x = (curMateCnt * newPosMatrix->x + ( matePos->x - mateOptPos->x ) ) / ( curMateCnt + 1.0f );
  newPosMatrix->y = (curMateCnt * newPosMatrix->y + ( matePos->y - mateOptPos->y ) ) / ( curMateCnt + 1.0f );
  newPosMatrix->z = (curMateCnt * newPosMatrix->z + ( matePos->z - mateOptPos->z ) ) / ( curMateCnt + 1.0f );
}

////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

__device__ float StabilCelluleMateBase( float3 * posIn, CellAttribute * attr, CrystalMatrixType crStr,
                                        int dir, int mateId, int nearInterstitial, int curDataId, int dataCount,
                                        /*short curTempr, */unsigned long curMask, float curMateCnt,
                                        float3 * curPos, float3 * newPosMatrix, bool extMate){

  int           mateType, mateDataId;
  unsigned long mateMask;
  float3        matePos, mateOptPos;

  mateType  = attr[mateId].type;
  mateMask  = attr[mateId].mask;
  __syncthreads();

  matePos.x = posIn[mateId].x;
  matePos.y = posIn[mateId].y;
  matePos.z = posIn[mateId].z;
  __syncthreads();

  if(__IS_CELL_MASK_BUG_INT_ANY(curMask)){
    if(extMate)return 0.0f;
    newPosMatrix->x = (curMateCnt * newPosMatrix->x + matePos.x) / (curMateCnt + 1.0f);
    newPosMatrix->y = (curMateCnt * newPosMatrix->y + matePos.y) / (curMateCnt + 1.0f);
    newPosMatrix->z = (curMateCnt * newPosMatrix->z + matePos.z) / (curMateCnt + 1.0f);
    return 1.0f;
  }
  __syncthreads();

  if(nearInterstitial != -1) {
    if(!__IS_CELL_MASK_BUG_INT_ANY(mateMask)){
      if(crStr == BCC){
        SetLaticePositionsBCC2(&mateOptPos, curPos, dir, nearInterstitial);
        StabilCelluleBase(curMateCnt, newPosMatrix, &matePos, &mateOptPos);
        return 1.0f;
      }
      if(crStr == FCC){
        SetLaticePositionsFCC2(&mateOptPos, curPos, dir, nearInterstitial);
        StabilCelluleBase(curMateCnt, newPosMatrix, &matePos, &mateOptPos);
        return 1.0f;
      }
      return 0.0f;
    }
    else return 0.0f;
  }
  __syncthreads();

  if(__IS_CELL_MASK_NEAR_BUG(curMask)){
    if(__IS_CELL_MASK_BUG_VAC(mateMask))        mateDataId = GetCalcDataIDX(true,  /*curTempr, cellData[curDataId].id,*/ -1, -1, crStr, dataCount);
    else if(__IS_CELL_MASK_BUG_SUB(mateMask))   mateDataId = GetCalcDataIDX(false, /*curTempr, cellData[curDataId].id,*/ mateType, -1, crStr, dataCount);
    else                                        mateDataId = curDataId;
    __syncthreads();
    if(mateDataId != -1){
      if(crStr == BCC){
        SetLaticePositionsBCC2(&mateOptPos, curPos, dir, mateDataId);
        StabilCelluleBase(curMateCnt, newPosMatrix, &matePos, &mateOptPos);
        return 1.0f;
      }
      __syncthreads();
      if(crStr == FCC){
        SetLaticePositionsFCC2(&mateOptPos, curPos, dir, mateDataId);
        StabilCelluleBase(curMateCnt, newPosMatrix, &matePos, &mateOptPos);
        return 1.0f;
      }
      __syncthreads();
      return 0.0f;
    }
    else return 0.0f;
  }
  __syncthreads();

  if(crStr == BCC){
    SetLaticePositionsBCC2(&mateOptPos, curPos, dir, curDataId);
    StabilCelluleBase(curMateCnt, newPosMatrix, &matePos, &mateOptPos);
    return 1.0f;
  }

  if(crStr == FCC){
    SetLaticePositionsBCC2(&mateOptPos, curPos, dir, curDataId);
    StabilCelluleBase(curMateCnt, newPosMatrix, &matePos, &mateOptPos);
    return 1.0f;
  }
  __syncthreads();
  return 0.0f;
}

////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

extern "C" __global__ void StabilisationBase( float * stressOut, float3 * posOut, float3 * posIn,
                                              CellAttribute * attr, CellMate * mate/*, short * tempr*/, CalcDataF * data,
                                              int cellCount, int dataCount, int steps){

  int idx = blockIdx.x * blockDim.x + threadIdx.x;
  int i, j;
  int cntLoops, cntMates;
  int mateId, curDataId, nearInterstitial;
  //short curTempr;
  CrystalMatrixType crStr;
  CellAttribute curAttr;
  float  stressMax = 0.0f, curMateCnt, force;
  float3 curPos, newPosMatrix;

  if(idx < dataCount)StoreCalcDataF(idx, data);
  __syncthreads();

  for(i = idx * steps ; (i < cellCount && i < (idx + 1) * steps); i++){
    curAttr.type      = attr[i].type;
    //curAttr.grainIdx  = attr[i].grainIdx;
    curAttr.mask      = attr[i].mask;
    __syncthreads();
//    curTempr          = tempr[i];
//    __syncthreads();

    crStr = CellMaskGetCrystalStruct(curAttr.mask);

    curDataId = GetCalcDataIDX(__IS_CELL_MASK_BUG_VAC(curAttr.mask), /*curTempr, curAttr.grainIdx,*/
                               __IS_CELL_MASK_BUG_SUB(curAttr.mask)      ? curAttr.type : -1,
                               __IS_CELL_MASK_BUG_INT_ANY(curAttr.mask)  ? curAttr.type : -1,
                               crStr, dataCount);
    __syncthreads();

    if(curDataId != -1 && curAttr.mask != CELL_MASK_NULL && !__IS_CELL_MASK_SCENE_BORDER(curAttr.mask)){

      curMateCnt      = 0.0f;
      newPosMatrix.x  = 0.0f; newPosMatrix.y  = 0.0f; newPosMatrix.z  = 0.0f;
      __syncthreads();

      curPos.x = posIn[i].x;
      curPos.y = posIn[i].y;
      curPos.z = posIn[i].z;
      __syncthreads();

      if(__IS_CELL_MASK_BUG_INT_ANY(curAttr.mask)){
        if(crStr == BCC){                           cntLoops = 6; }
        else if(crStr == FCC){
          if(6 == CellMaskGetLattice(curAttr.mask)) cntLoops = 4;
          else                                      cntLoops = 6;
        } else {                                    cntLoops = 0; }
      }
      else if(crStr == BCC){                        cntLoops = CNT_DIR_BCC; }
      else if(crStr == FCC){                        cntLoops = CNT_DIR_FCC; }
      else{                                         cntLoops = 0; }
      __syncthreads();

      nearInterstitial = -1;
      if(__IS_CELL_MASK_NEAR_BUG(curAttr.mask)){
        for(j = 0 ; j < cntLoops ; j++){
          mateId = mate[i].m1[j];                                                                                             __syncthreads();
          if( mateId != 0)nearInterstitial = IsNearInterstitial(attr, mateId, /*curTempr, curAttr.grainIdx,*/ crStr, dataCount);  __syncthreads();
          if(-1 != nearInterstitial)break;                                                                                    __syncthreads();
          mateId = mate[i].m2[j];                                                                                             __syncthreads();
          if( mateId != 0)nearInterstitial = IsNearInterstitial(attr, mateId, /*curTempr, curAttr.grainIdx,*/ crStr, dataCount);  __syncthreads();
          if(-1 != nearInterstitial)break;                                                                                    __syncthreads();
        }
        __syncthreads();
      }
      __syncthreads();

      for(j = 0 ; j < cntLoops ; j++){
        cntMates = CellMaskGetCntMates(curAttr.mask, j);
        __syncthreads();

        if(cntMates > 0){
          mateId = mate[i].m1[j];  __syncthreads();
          if( mateId != 0){
            curMateCnt += StabilCelluleMateBase(posIn, attr, crStr, j, mateId, nearInterstitial, curDataId, dataCount/*, curTempr*/, curAttr.mask, curMateCnt, &curPos, &newPosMatrix, false);
            __syncthreads();
          }
        }
        __syncthreads();

        if(cntMates > 1){
          mateId = mate[i].m2[j]; __syncthreads();
          if( mateId != 0){
            curMateCnt += StabilCelluleMateBase(posIn, attr, crStr, j, mateId, nearInterstitial, curDataId, dataCount/*, curTempr*/, curAttr.mask, curMateCnt, &curPos, &newPosMatrix, true);
            __syncthreads();
          }
        }
        __syncthreads();

      }
      __syncthreads();
    }
    __syncthreads();

    if(!__IS_CELL_MASK_BUG_INT_ANY(curAttr.mask)){
      newPosMatrix.x = curPos.x + newPosMatrix.x;
      newPosMatrix.y = curPos.y + newPosMatrix.y;
      newPosMatrix.z = curPos.z + newPosMatrix.z;
    }
    __syncthreads();
    posOut[i].x = newPosMatrix.x;
    posOut[i].y = newPosMatrix.y;
    posOut[i].z = newPosMatrix.z;
    __syncthreads();

    newPosMatrix.x -= curPos.x;
    newPosMatrix.y -= curPos.y;
    newPosMatrix.z -= curPos.z;
    force = sqrt(newPosMatrix.x*newPosMatrix.x + newPosMatrix.y*newPosMatrix.y + newPosMatrix.z*newPosMatrix.z);
    __syncthreads();

    if(stressMax < force) stressMax = force;
    __syncthreads();

  }
  __syncthreads();
  stressOut[idx] = stressMax;
  __syncthreads();
}

#endif
