#ifndef ___PROCES_F_STABIL_LJ_2_H___
#define ___PROCES_F_STABIL_LJ_2_H___

#include "proces_f_cell_data.h"

#define _1_DIV_6_   0.16666667f

__device__ float ReverseLJP(float y, float A, float B, float d){

  float T, DIS;

  DIS = B*B + 4.0f * A * y;
  if(DIS < 0.0f)return 0.0f;

  DIS = sqrt(DIS);

  T = (-B + DIS) / (2.0f*y);
  if(T > 0.0f){
    T = pow(T, _1_DIV_6_);
    return d - T;
  }

  T = (-B - DIS) / (2.0f*y);
  if(T > 0.0f){
    T = pow(T, _1_DIV_6_);
    return d - T;
  }

  return 0.0f;
}

__device__ float ReverseLJM(float y, float A, float B, float d){

  float T, DIS;

  DIS = B*B - 4.0f * A * y;
  if(DIS < 0.0f)return 0.0f;

  DIS = sqrt(DIS);

  T = (-B + DIS) / (-2.0f*y);
  if(T > 0.0f){
    T = pow(T, _1_DIV_6_);
    return T - d;
  }

  T = (-B - DIS) / (-2.0f*y);
  if(T > 0.0f){
    T = pow(T, _1_DIV_6_);
    return T - d;
  }

  return 0.0f;
}

__device__ void StabilCelluleLJ2(float3 * forceLJP, float3 * forceLJM, float3 * curPos, float3 * matePos, int mateDataId){

  float3 tmp;
  float  force;

  tmp.x = curPos->x - matePos->x;
  tmp.y = curPos->y - matePos->y;
  tmp.z = curPos->z - matePos->z;

  force = sqrt(tmp.x*tmp.x + tmp.y*tmp.y + tmp.z*tmp.z);
  if(force != 0.0f){
    tmp.x /= force;
    tmp.y /= force;
    tmp.z /= force;

    force = force * force;
    force = force * force * force;
    force = cellData[mateDataId].lA/(force*force) - cellData[mateDataId].lB/force;

    tmp.x *= force;
    tmp.y *= force;
    tmp.z *= force;

    if(force > 0.0f){
      forceLJP->x += tmp.x;
      forceLJP->y += tmp.y;
      forceLJP->z += tmp.z;
    } else {
      forceLJM->x += tmp.x;
      forceLJM->y += tmp.y;
      forceLJM->z += tmp.z;
    }
  }
}

////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

__device__ float StabilCelluleMateLJ2(float3 * posIn, CellAttribute * attr, CrystalStructure crStr,
                                   int mateId, int nearInterstitial, int curDataId, int dataCount,
                                   short curTempr, unsigned long curMask, float curMateCnt,
                                   float3 * curPos, float3 * forceLJP, float3 * forceLJM, bool extMate){

  int           mateType, mateDataId;
  unsigned long mateMask;
  float3        matePos;

  mateType  = attr[mateId].type;
  mateMask  = attr[mateId].mask;
  __syncthreads();

  matePos.x = posIn[mateId].x;
  matePos.y = posIn[mateId].y;
  matePos.z = posIn[mateId].z;
  __syncthreads();

  if(__IS_CELL_MASK_BUG_INT_ANY(curMask)){
    if(extMate)return 0.0f;
    forceLJP->x = (curMateCnt * forceLJP->x + matePos.x) / (curMateCnt + 1.0f);
    forceLJP->y = (curMateCnt * forceLJP->y + matePos.y) / (curMateCnt + 1.0f);
    forceLJP->z = (curMateCnt * forceLJP->z + matePos.z) / (curMateCnt + 1.0f);
    return 1.0f;
  }
  __syncthreads();

  if(nearInterstitial != -1) {
    if(!__IS_CELL_MASK_BUG_INT_ANY(mateMask)){
      if(crStr == BCC__CrystalStructure){
        StabilCelluleLJ2(forceLJP, forceLJM, curPos, &matePos, nearInterstitial);
        return 1.0f;
      }
      if(crStr == FCC__CrystalStructure){
        StabilCelluleLJ2(forceLJP, forceLJM, curPos, &matePos, nearInterstitial);
        return 1.0f;
      }
      return 0.0f;
    }
    else return 0.0f;
  }
  __syncthreads();

  if(__IS_CELL_MASK_NEAR_BUG(curMask)){
    if(__IS_CELL_MASK_BUG_VAC(mateMask))        mateDataId = GetCalcDataIDX(true,  curTempr, cellData[curDataId].id, -1, -1, crStr, dataCount);
    else if(__IS_CELL_MASK_BUG_SUB(mateMask))   mateDataId = GetCalcDataIDX(false, curTempr, cellData[curDataId].id, mateType, -1, crStr, dataCount);
    else                                        mateDataId = curDataId;
    __syncthreads();
    if(mateDataId != -1){
      if(crStr == BCC__CrystalStructure){
        StabilCelluleLJ2(forceLJP, forceLJM, curPos, &matePos, mateDataId);
        return 1.0f;
      }
      __syncthreads();
      if(crStr == FCC__CrystalStructure){
        StabilCelluleLJ2(forceLJP, forceLJM, curPos, &matePos, mateDataId);
        return 1.0f;
      }
      __syncthreads();
      return 0.0f;
    }
    else return 0.0f;
  }
  __syncthreads();

  if(crStr == BCC__CrystalStructure){
    StabilCelluleLJ2(forceLJP, forceLJM, curPos, &matePos, curDataId);
    return 1.0f;
  }

  if(crStr == FCC__CrystalStructure){
    StabilCelluleLJ2(forceLJP, forceLJM, curPos, &matePos, curDataId);
    return 1.0f;
  }
  __syncthreads();
  return 0.0f;
}

////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

extern "C" __global__ void StabilisationLJ2(  float * stressOut, float3 * posOut, float3 * posIn,
                                              CellAttribute * attr, CellMate * mate, short * tempr, CalcDataF * data,
                                              int cellCount, int dataCount, int steps){

  int idx = blockIdx.x * blockDim.x + threadIdx.x;
  int i, j;
  int cntLoops, cntMates;
  int mateId, curDataId, nearInterstitial;
  short curTempr;
  CrystalStructure crStr;
  CellAttribute curAttr;
  float  stressMax = 0.0f, curMateCnt, force;
  float3 curPos, forceLJP, forceLJM;

  if(idx < dataCount)StoreCalcDataF(idx, data);
  __syncthreads();

  for(i = idx * steps ; (i < cellCount && i < (idx + 1) * steps); i++){
    curAttr.type      = attr[i].type;
    curAttr.grainIdx  = attr[i].grainIdx;
    curAttr.mask      = attr[i].mask;
    __syncthreads();
    curTempr          = tempr[i];
    __syncthreads();

    crStr = CellMaskGetCrystalStruct(curAttr.mask);

    curDataId = GetCalcDataIDX(__IS_CELL_MASK_BUG_VAC(curAttr.mask), curTempr, curAttr.grainIdx,
                               __IS_CELL_MASK_BUG_SUB(curAttr.mask)      ? curAttr.type : -1,
                               __IS_CELL_MASK_BUG_INT_ANY(curAttr.mask)  ? curAttr.type : -1,
                               crStr, dataCount);
    __syncthreads();

    if(curDataId != -1 && curAttr.mask != CELL_MASK_NULL && !__IS_CELL_MASK_SCENE_BORDER(curAttr.mask)){

      curMateCnt = 0.0f;
      forceLJP.x      = 0.0f; forceLJP.y      = 0.0f; forceLJP.z      = 0.0f;
      forceLJM.x      = 0.0f; forceLJM.y      = 0.0f; forceLJM.z      = 0.0f;
      __syncthreads();

      curPos.x = posIn[i].x;
      curPos.y = posIn[i].y;
      curPos.z = posIn[i].z;
      __syncthreads();

      if(__IS_CELL_MASK_BUG_INT_ANY(curAttr.mask)){
        if(crStr == BCC__CrystalStructure){         cntLoops = 6; }
        else if(crStr == FCC__CrystalStructure){
          if(6 == CellMaskGetLattice(curAttr.mask)) cntLoops = 4;
          else                                      cntLoops = 6;
        } else {                                    cntLoops = 0; }
      }
      else if(crStr == BCC__CrystalStructure){      cntLoops = CNT_DIR_BCC; }
      else if(crStr == FCC__CrystalStructure){      cntLoops = CNT_DIR_FCC; }
      else{                                         cntLoops = 0; }
      __syncthreads();

      nearInterstitial = -1;
      if(__IS_CELL_MASK_NEAR_BUG(curAttr.mask)){
        for(j = 0 ; j < cntLoops ; j++){
          mateId = mate[i].m1.mate[j];                                                                                        __syncthreads();
          if( mateId != 0)nearInterstitial = IsNearInterstitial(attr, mateId, curTempr, curAttr.grainIdx, crStr, dataCount);  __syncthreads();
          if(-1 != nearInterstitial)break;                                                                                    __syncthreads();
          mateId = mate[i].m2.mate[j];                                                                                        __syncthreads();
          if( mateId != 0)nearInterstitial = IsNearInterstitial(attr, mateId, curTempr, curAttr.grainIdx, crStr, dataCount);  __syncthreads();
          if(-1 != nearInterstitial)break;                                                                                    __syncthreads();
          mateId = mate[i].m3.mate[j];                                                                                        __syncthreads();
          if( mateId != 0)nearInterstitial = IsNearInterstitial(attr, mateId, curTempr, curAttr.grainIdx, crStr, dataCount);  __syncthreads();
          if(-1 != nearInterstitial)break;                                                                                    __syncthreads();
        }
        __syncthreads();
      }
      __syncthreads();

      for(j = 0 ; j < cntLoops ; j++){
        cntMates = CellMaskGetCntMates(curAttr.mask, j);
        __syncthreads();

        if(cntMates > 0){
          mateId = mate[i].m1.mate[j];  __syncthreads();
          if( mateId != 0){
            curMateCnt += StabilCelluleMateLJ2(posIn, attr, crStr, mateId, nearInterstitial, curDataId, dataCount, curTempr, curAttr.mask, curMateCnt, &curPos, &forceLJP, &forceLJM, false);
            __syncthreads();
          }
        }
        __syncthreads();

        if(cntMates > 1){
          mateId = mate[i].m2.mate[j]; __syncthreads();
          if( mateId != 0){
            curMateCnt += StabilCelluleMateLJ2(posIn, attr, crStr, mateId, nearInterstitial, curDataId, dataCount, curTempr, curAttr.mask, curMateCnt, &curPos, &forceLJP, &forceLJM, true);
            __syncthreads();
          }
        }
        __syncthreads();

        if(cntMates > 2){
          mateId = mate[i].m3.mate[j]; __syncthreads();
          if( mateId != 0){
            curMateCnt += StabilCelluleMateLJ2(posIn, attr, crStr, mateId, nearInterstitial, curDataId, dataCount, curTempr, curAttr.mask, curMateCnt, &curPos, &forceLJP, &forceLJM, true);
            __syncthreads();
          }
        }
        __syncthreads();
      }
      __syncthreads();
    }
    __syncthreads();

    if(     crStr == BCC__CrystalStructure){ curMateCnt = (float)CNT_DIR_BCC; }
    else if(crStr == FCC__CrystalStructure){ curMateCnt = (float)CNT_DIR_FCC; }
    else{                                    curMateCnt = (float)MAX_MATES;   }

    if(!__IS_CELL_MASK_BUG_INT_ANY(curAttr.mask)){

      force = sqrt(forceLJP.x*forceLJP.x + forceLJP.y*forceLJP.y + forceLJP.z*forceLJP.z);
      if(curMateCnt != 0.0f)force /= curMateCnt;
      __syncthreads();
      if(force == 0.0f){
        forceLJP.x = 0.0f;
        forceLJP.y = 0.0f;
        forceLJP.z = 0.0f;
      }else{
        forceLJP.x /= force;
        forceLJP.y /= force;
        forceLJP.z /= force;
        force = ReverseLJP(force, cellData[curDataId].lA, cellData[curDataId].lB, cellData[curDataId].d);
        forceLJP.x *= force;
        forceLJP.y *= force;
        forceLJP.z *= force;
      }
      __syncthreads();

      force = sqrt(forceLJM.x*forceLJM.x + forceLJM.y*forceLJM.y + forceLJM.z*forceLJM.z);
      if(curMateCnt != 0.0f)force /= curMateCnt;
      __syncthreads();
      if(force == 0.0f){
        forceLJM.x = 0.0f;
        forceLJM.y = 0.0f;
        forceLJM.z = 0.0f;
      }else{
        forceLJM.x /= force;
        forceLJM.y /= force;
        forceLJM.z /= force;
        force = ReverseLJM(force, cellData[curDataId].lA, cellData[curDataId].lB, cellData[curDataId].d);
        forceLJM.x *= force;
        forceLJM.y *= force;
        forceLJM.z *= force;
      }
      __syncthreads();

      forceLJM.x += forceLJP.x;
      forceLJM.y += forceLJP.y;
      forceLJM.z += forceLJP.z;

      forceLJP.x = curPos.x + forceLJM.x;
      forceLJP.y = curPos.y + forceLJM.y;
      forceLJP.z = curPos.z + forceLJM.z;
      force = sqrt(forceLJM.x*forceLJM.x + forceLJM.y*forceLJM.y + forceLJM.z*forceLJM.z);
    } else {
      forceLJM.x = curPos.x - forceLJP.x;
      forceLJM.y = curPos.y - forceLJP.y;
      forceLJM.z = curPos.z - forceLJP.z;
      force = sqrt(forceLJM.x*forceLJM.x + forceLJM.y*forceLJM.y + forceLJM.z*forceLJM.z);
    }
    __syncthreads();

    posOut[i].x = forceLJP.x;
    posOut[i].y = forceLJP.y;
    posOut[i].z = forceLJP.z;
    __syncthreads();

    if(stressMax < force) stressMax = force;
    __syncthreads();
  }
  __syncthreads();
  stressOut[idx] = stressMax;
  __syncthreads();
}

#endif
