#ifndef ___PROCES_F_STATISTICS_H__
#define ___PROCES_F_STATISTICS_H__

#ifndef CUDA_KERNEL_INCLUDE
  #define CUDA_KERNEL_INCLUDE
  #define QT_VERSION
#endif

#include "../gpu_structs.h"
#include <vector_functions.h>
#include <float.h>

extern "C" __global__ void Statistics(StatsDataF * out, float3 * stress, CellAttribute * attr, short * tempr, int cellCount, int steps){

  int idx = blockIdx.x * blockDim.x + threadIdx.x;
  int i;

  short         curTempr;
  float         stresAbs;
  float3        tmp;
  CellAttribute curAttr;
  StatsDataF    tmpData;

  tmpData.minTemperature    = (short)30000;
  tmpData.maxTemperature    = (short)0;
  tmpData.cellCount         = 0;
  tmpData.matesCount        = 0;
  tmpData.laticeCount       = 0;
  tmpData.vacCount          = 0;
  tmpData.intCount          = 0;
  tmpData.subCount          = 0;
  tmpData.minTemperatureId  = 0;
  tmpData.maxTemperatureId  = 0;
  tmpData.minStresId        = 0;
  tmpData.maxStresId        = 0;
  tmpData.minForceId        = 0;
  tmpData.maxForceId        = 0;
  tmpData.avgTemperature    = 0.0f;
  tmpData.minStresAbs       = FLT_MAX;
  tmpData.maxStresAbs       = FLT_MIN;
  tmpData.avgStresAbs       = 0.0f;
  tmpData.minForce          = FLT_MAX;
  tmpData.maxForce          = FLT_MIN;
  tmpData.avgForce          = 0.0f;
  tmpData.systemForce       = 0.0f;
  tmpData.minStresX = 0.0f; tmpData.minStresY = 0.0f;   tmpData.minStresZ = 0.0f;
  tmpData.maxStresX = 0.0f; tmpData.maxStresY = 0.0f;   tmpData.maxStresZ = 0.0f;
  tmpData.avgStresX = 0.0f; tmpData.avgStresY = 0.0f;   tmpData.avgStresZ = 0.0f;
  __syncthreads();

  for(i = idx * steps ; (i < cellCount && i < (idx + 1) * steps); i++){
    curAttr.mask      = attr[i].mask;
    __syncthreads();

    if(curAttr.mask != CELL_MASK_NULL){
      curAttr.type      = attr[i].type;
      curAttr.grainIdx  = attr[i].grainIdx;
      __syncthreads();

      curTempr          = tempr[i];
      __syncthreads();

      tmp.x = stress[i].x;
      tmp.y = stress[i].y;
      tmp.z = stress[i].z;
      __syncthreads();

      tmpData.cellCount++;

      stresAbs = sqrt(tmp.x*tmp.x + tmp.y*tmp.y + tmp.z*tmp.z);

      tmpData.avgTemperature  += (float)curTempr;
      tmpData.avgStresAbs     += stresAbs;
      tmpData.avgStresX       += tmp.x;
      tmpData.avgStresY       += tmp.y;
      tmpData.avgStresZ       += tmp.z;

      if(tmpData.minTemperature > curTempr){ tmpData.minTemperature = curTempr; tmpData.minTemperatureId = i; }             __syncthreads();
      if(tmpData.maxTemperature < curTempr){ tmpData.maxTemperature = curTempr; tmpData.maxTemperatureId = i; }             __syncthreads();
      if(tmpData.laticeCount < CellMaskGetLattice(curAttr.mask)){ tmpData.laticeCount = CellMaskGetLattice(curAttr.mask); } __syncthreads();

      if(__IS_CELL_MASK_BUG_VAC(curAttr.mask))tmpData.vacCount++;     __syncthreads();
      if(__IS_CELL_MASK_BUG_SUB(curAttr.mask))tmpData.subCount++;     __syncthreads();
      if(__IS_CELL_MASK_BUG_INT_ANY(curAttr.mask))tmpData.intCount++; __syncthreads();

      if(tmpData.minStresAbs > stresAbs){
        tmpData.minStresAbs = stresAbs;
        tmpData.minStresId  = i;
        tmpData.minStresX   = tmp.x;
        tmpData.minStresY   = tmp.y;
        tmpData.minStresZ   = tmp.z;
      }
      __syncthreads();

      if(tmpData.maxStresAbs < stresAbs){
        tmpData.maxStresAbs = stresAbs;
        tmpData.maxStresId  = i;
        tmpData.maxStresX   = tmp.x;
        tmpData.maxStresY   = tmp.y;
        tmpData.maxStresZ   = tmp.z;
      }
      __syncthreads();

      tmpData.matesCount += CellMaskGetCntMatesAll(curAttr.mask);
      __syncthreads();

    }
    __syncthreads();
  }
  __syncthreads();

  stresAbs = (float)tmpData.cellCount;
  if(stresAbs != 0.0f){
    tmpData.avgTemperature  /= stresAbs;
    tmpData.avgStresAbs     /= stresAbs;
    tmpData.avgStresX       /= stresAbs;
    tmpData.avgStresY       /= stresAbs;
    tmpData.avgStresZ       /= stresAbs;
  }
  __syncthreads();

//  stresAbs = (float)tmpData.matesCount;
//  tmpData.avgForce = tmpData.systemForce / stresAbs;
//  __syncthreads();

  if(tmpData.maxForce == FLT_MIN)tmpData.maxForce = 0.0f;
  if(tmpData.minForce == FLT_MAX)tmpData.minForce = 0.0f;
  __syncthreads();

  out[idx].minTemperature   = tmpData.minTemperature;
  out[idx].maxTemperature   = tmpData.maxTemperature;
  out[idx].cellCount        = tmpData.cellCount;
  out[idx].matesCount       = tmpData.matesCount;
  out[idx].laticeCount      = tmpData.laticeCount;
  out[idx].vacCount         = tmpData.vacCount;
  out[idx].intCount         = tmpData.intCount;
  out[idx].subCount         = tmpData.subCount;
  out[idx].minTemperatureId = tmpData.minTemperatureId;
  out[idx].maxTemperatureId = tmpData.maxTemperatureId;
  out[idx].minStresId       = tmpData.minStresId;
  out[idx].maxStresId       = tmpData.maxStresId;
  out[idx].minForceId       = tmpData.minForceId;
  out[idx].maxForceId       = tmpData.maxForceId;
  out[idx].avgTemperature   = tmpData.avgTemperature;
  out[idx].minStresAbs      = tmpData.minStresAbs;
  out[idx].maxStresAbs      = tmpData.maxStresAbs;
  out[idx].avgStresAbs      = tmpData.avgStresAbs;
  out[idx].minForce         = tmpData.minForce;
  out[idx].maxForce         = tmpData.maxForce;
  out[idx].avgForce         = tmpData.avgForce;
  out[idx].systemForce      = tmpData.systemForce;
  out[idx].minStresX        = tmpData.minStresX;
  out[idx].minStresY        = tmpData.minStresY;
  out[idx].minStresZ        = tmpData.minStresZ;
  out[idx].maxStresX        = tmpData.maxStresX;
  out[idx].maxStresY        = tmpData.maxStresY;
  out[idx].maxStresZ        = tmpData.maxStresZ;
  out[idx].avgStresX        = tmpData.avgStresX;
  out[idx].avgStresY        = tmpData.avgStresY;
  out[idx].avgStresZ        = tmpData.avgStresZ;
  __syncthreads();
}

#endif // ___CUDA_STATISTICS_CU__
