#include "project.h"

ProjectDataF::ProjectDataF(int _count, int _seedIdx) : count(_count), countVac(0), seedIdx(_seedIdx), countBorder(0){
  if(count > 0){
    pos  = new VecF[count];
    stres= new VecF[count];
    attr = new CellAttribute[count];
    mate = new CellMate[count];
    memset(pos,  0, count * sizeof(VecF));
    memset(stres,0, count * sizeof(VecF));
    memset(attr, 0, count * sizeof(CellAttribute));
    memset(mate, 0, count * sizeof(CellMate));
  }else{
    pos  = 0;
    stres= 0;
    attr = 0;
    mate = 0;
  }
  wallIds = 0;
}

ProjectDataF::~ProjectDataF() {
  if(pos      != 0) delete pos;
  if(stres    != 0) delete stres;
  if(attr     != 0) delete attr;
  if(mate     != 0) delete mate;
  if(wallIds  != 0) delete wallIds;
}

bool ProjectDataF::storeMol2File(QString fileNameBase, QString name, int stepId, bool storeBonds)
{
  if(count <= 0)return false;
  QString filename = fileNameBase + "_" + QString::number(stepId) + ".mol2";

  QFile file(filename);
  if (!file.open(QIODevice::WriteOnly | QIODevice::Text)) return false;
  QTextStream out(&file);

  int * idxMap = new int[count];
  idxMap[0] = 0;
  int cellId = 1;
  for(int i = 1 ; i < count ; i++){
    if(__IS_CELL_MASK_BUG_VAC(attr[i].mask))idxMap[i] = 0;
    else { idxMap[i] = cellId; cellId++; }
  }

  int countBonds = 0;
  if(storeBonds){
    for(int i = 1 ; i < count ; i++){
      if(__IS_CELL_MASK_BUG_VAC(attr[i].mask))continue;
      for(int j = 0 ; j < MAX_MATES ; j++){
        if(i < mate[i].m1[j] && !__IS_CELL_MASK_BUG_VAC(attr[mate[i].m1[j]].mask))countBonds++;
        if(i < mate[i].m2[j] && !__IS_CELL_MASK_BUG_VAC(attr[mate[i].m2[j]].mask))countBonds++;
      }
    }
  }
  out << "@<TRIPOS>MOLECULE\n";
  out << "Generated by Atomic Networks[step=" << stepId << "]: " << qPrintable(name) << "\n";
  out << (count - 1 - countVac) << " " << countBonds << "\n";
  out << "NO_CHARGES\n";
  out << "@<TRIPOS>ATOM\n";

  cellId = 1;
  for(int i = 1 ; i < count ; i++){
    if(__IS_CELL_MASK_BUG_VAC(attr[i].mask))continue;
    out << cellId++ << " A " << pos[i].x << " " << pos[i].y << " " << pos[i].z << " A\n";
  }
  if(storeBonds){
    out << "@<TRIPOS>BOND\n";
    cellId = 1;
    int bondId = 1;
    for(int i = 1 ; i < count ; i++){
      if(__IS_CELL_MASK_BUG_VAC(attr[i].mask))continue;
      for(int j = 0 ; j < MAX_MATES ; j++){
        if(i < mate[i].m1[j] && !__IS_CELL_MASK_BUG_VAC(attr[mate[i].m1[j]].mask))
          out << bondId++ << " " << cellId << " " << idxMap[mate[i].m1[j]] << " 1\n";
        if(i < mate[i].m2[j] && !__IS_CELL_MASK_BUG_VAC(attr[mate[i].m2[j]].mask))
          out << bondId++ << " " << cellId << " " << idxMap[mate[i].m2[j]] << " 1\n";
      }
      cellId++;
    }
  }
  out.flush();
  file.flush();
  file.close();
  return true;
}

ProjectF::ProjectF() : wallX0(0.0f), wallX1(0.0f), wallY0(0.0f), wallY1(10.0f), wallZ0(10.0f), wallZ1(10.0f), wallMoveSteps(2), wallMoveLength(0.001f),
  crystalA(1.0f), crystalLJE(1.0f), crystalLJD(0.866f), useVacancy(false), vacancyRatio(0.0f), useSubstitution(false), substitutionRatio(0.0f), substitutionLJE(0.0f), substitutionLJD(0.0f)

{
  wallMoveType = PRESS_X;
  matrixType = BCC;
  seed.x = (wallX1 - wallX0) / 2.0f;
  seed.y = (wallY1 - wallY0) / 2.0f;
  seed.z = (wallZ1 - wallZ0) / 2.0f;
  alpha.x = 1.0; alpha.y = 0.0; alpha.z = 0.0;
  beta.x  = 0.0; beta.y  = 1.0; beta.z  = 0.0;
  gama.x  = 0.0; gama.y  = 0.0; gama.z  = 1.0;
}
